#include "amxxmodule.h"

#ifdef __linux__
	#include <sys/mman.h>
#endif

// Define memory address type
#ifdef __amd64__
	typedef uint64_t maddress;
#else
	typedef uint32_t maddress;
#endif
	
#define BYTE_BYTES 1
// Return codes for MemoryProtect
#define MP_FAIL -1
#define MP_OK   0
// Change name Status
#define NAME_CHANGING 0
#define NAME_CHANGED 1

// Memory protection constants
#ifdef __linux__
	#define MPROT_CODE        PROT_READ|PROT_EXEC
	#define MPROT_DATA        PROT_READ|PROT_WRITE
	#define MPROT_RODATA      PROT_READ
	#define MPROT_CODE_EDIT   PROT_READ|PROT_WRITE|PROT_EXEC
	#define MPROT_RODATA_EDIT PROT_READ|PROT_WRITE
#else
	#define MPROT_CODE        PAGE_EXECUTE_READ
	#define MPROT_DATA        PAGE_READWRITE
	#define MPROT_RODATA      PAGE_READONLY
	#define MPROT_CODE_EDIT   PAGE_EXECUTE_READWRITE
	#define MPROT_RODATA_EDIT PAGE_READWRITE
#endif

// Memory area types
#define MEMTYPE_CODE   0 // Code (usually .text segment, requires mprotect or VirtualProtect)
#define MEMTYPE_DATA   1 // Data (usually .data segment, writable by default)
#define MEMTYPE_RODATA 2 // Read-Only Data (usually .rodata on Linux, .rdata on Windows)

#define SAMPLE_DLLFUNC reinterpret_cast<void*>(gpGamedllFuncs->dllapi_table->pfnThink)
#define SAMPLE_ENGFUNC reinterpret_cast<void*>(g_engfuncs.pfnChangeLevel)

maddress gameDllAddress;
maddress gameEngAddress;

int MemoryProtect(void *addr, size_t len, unsigned long newProt, unsigned long *oldProt, char memType = MEMTYPE_CODE);
int UTIL_PatchMemory_UnsignedByte(maddress baseaddress, maddress address, unsigned char patch, char memType = MEMTYPE_DATA);
maddress GetRealMemoryAddress(maddress baseaddress,maddress address, char memType);
bool GetBaseAddress(void *pAddr, maddress &pBaseAddr);
bool GetBaseAddresses( void );

bool IsDuplicated(int index, char *name);

bool status[33][2];
char newname[33][32];

static cell AMX_NATIVE_CALL set_utf8_name(AMX *amx, cell *params)
{
	int index = params[1];
	
	if (index < 0 || index > gpGlobals->maxEntities)
	{
		MF_LogError(amx, AMX_ERR_NATIVE, "Player out of range (%d)", index);
		return 0;
	}else{
		if (!MF_IsPlayerIngame(index) || FNullEnt(MF_GetPlayerEdict(index)))
		{
			MF_LogError(amx, AMX_ERR_NATIVE, "Invalid player %d", index);
			return 0;
		}
	}
	edict_t *pEntity = MF_GetPlayerEdict(index);
	if (!MF_IsPlayerAlive(index))
	{
		g_engfuncs.pfnClientPrintf(pEntity, print_center, "only aliver can change the utf name.\n");
		return 0;
	}
	int len, i = 0;
	char *utf8name = MF_GetAmxString(amx, params[2], 0, &len);
	if (strchr(utf8name, '\\')
		|| strchr(utf8name, '\n')
		|| strchr(utf8name, '\r')
		|| strchr(utf8name, '\f')
		|| strchr(utf8name, '\t')
		|| strchr(utf8name, '\v'))
	{
		g_engfuncs.pfnClientPrintf(pEntity, print_center, "Invalid name.\n");
		return 0;
	}
	if (IsDuplicated(index, utf8name))
	{
		return 0;
	}
	cell cchar;
	for(;;)
	{
		cchar = utf8name[i++];
		if (cchar < 0)
		{
			status[index][NAME_CHANGING] = true;
			strcpy(newname[index], utf8name);
			char fakename[32] = "0123456789012345678901234567890";
			fakename[strlen(utf8name)] = 0;
			SET_CLIENT_KEYVALUE(index, GET_INFOKEYBUFFER(MF_GetPlayerEdict(index)), "name", "0123456789012345678901234567890");
			return 2;
		}
		else if (cchar == 0)
		{
			SET_CLIENT_KEYVALUE(index, GET_INFOKEYBUFFER(MF_GetPlayerEdict(index)), "name", utf8name);
			return 1;
		}
	}
	return 0;
}

bool IsDuplicated(int index, char *name)
{
	edict_t *pEntity = MF_GetPlayerEdict(index);
	for (int a = 1; a <= gpGlobals->maxEntities; a++)
	{
		if (!MF_IsPlayerIngame(a))continue;
		if (index != a && stricmp(name, newname[a]) == 0 && status[a][NAME_CHANGED])
		{
			g_engfuncs.pfnClientPrintf(pEntity, print_center, "Your name shouldn't be the same as other's.\n");
			return true;
		}
	}
	return false;
}
bool GetBaseAddress(void *pAddr, maddress &pBaseAddr)
{
#ifdef WIN32
	MEMORY_BASIC_INFORMATION mem;
	if (!VirtualQuery(pAddr, &mem, sizeof(mem)))
		return false;

	pBaseAddr = (maddress)mem.AllocationBase;
	
	IMAGE_DOS_HEADER *dos = (IMAGE_DOS_HEADER *)(mem.AllocationBase);
	IMAGE_NT_HEADERS *pe = reinterpret_cast<IMAGE_NT_HEADERS*>( (unsigned long)dos + (unsigned long)dos->e_lfanew );
	if (pe->Signature != IMAGE_NT_SIGNATURE)
		return false;

	//if (memLength)
		//*memLength = (size_t)(pe->OptionalHeader.SizeOfImage);
	return true;
#else
	Dl_info info;
	struct stat buf;
	
	if (!dladdr(pAddr, &info))
		return false;
	
	if (!info.dli_fbase || !info.dli_fname)
		return false;
	
	if (stat(info.dli_fname, &buf) != 0)
		return false;
	
	if (pBaseAddr)
		*pBaseAddr = (unsigned char *)info.dli_fbase;
	//if (memLength)
		//*memLength = buf.st_size;
	
	return true;
#endif
}

int MemoryProtect(void *addr, size_t len, unsigned long newProt, unsigned long *oldProt, char memType) {
	int retVal;

	#ifdef __linux__
		maddress alignAddr = (maddress)addr - ((maddress)addr % pageSize);
		retVal = mprotect((void*)alignAddr, pageSize, newProt);

		// Linux's mprotect doesn't get the old protection flags, so we have to fake it
		switch (memType) {
			case MEMTYPE_CODE:
				*oldProt = MPROT_CODE;
				break;
			case MEMTYPE_RODATA:
				*oldProt = MPROT_RODATA;
				break;
			default:
				*oldProt = MPROT_CODE;
				break;
		}
	#else
		retVal = VirtualProtect(addr, len, newProt, oldProt);
		// This will match the Windows return value with the Linux ones, done for consistency
		if (retVal == 0) {
			retVal = -1;
		} else {
			retVal = 0;
		}
	#endif

	return retVal;
}

// Linux won't work till I fix it for MEMTYPE_DATA
#ifdef __linux__
	// Data section stuff
	maddress dataSectionStart;
	maddress dataSectionOffset;

	int pageSize = sysconf(_SC_PAGESIZE);
#endif

/* Gets real memory address */
maddress GetRealMemoryAddress(maddress baseaddress, maddress address, char memType) 
{
	if(baseaddress == NULL) return address;
	maddress realAddress = address;
	switch (memType) 
	{
		case MEMTYPE_CODE: case MEMTYPE_RODATA:
			realAddress = baseaddress + address;
			break;
		case MEMTYPE_DATA:
			// Linux's data segment is in a not so simple place in memory
			#ifdef __linux__
				realAddress = dataSectionStart + (address - dataSectionOffset);
			#else
				realAddress = baseaddress + address;
			#endif
			break;
	}
	return realAddress;
}

int UTIL_PatchMemory_UnsignedByte(maddress baseaddress, maddress address, unsigned char patch, char memType)
{
	unsigned long oldProtect = 0;
	maddress realAddress = GetRealMemoryAddress(baseaddress, address, memType);

	switch (memType)
	{
	case MEMTYPE_CODE:
		{
			if (MemoryProtect((void*)realAddress, BYTE_BYTES, MPROT_CODE_EDIT, &oldProtect, memType) == MP_FAIL) return MP_FAIL;
			break;
		}
	case MEMTYPE_RODATA:
		{
			if (MemoryProtect((void*)realAddress, BYTE_BYTES, MPROT_RODATA_EDIT, &oldProtect, memType) == MP_FAIL) return MP_FAIL;
			break;
		}
	}
	*(unsigned char *)realAddress = patch;

	if (memType == MEMTYPE_CODE)
	{
		MemoryProtect((void*)realAddress, BYTE_BYTES, oldProtect, &oldProtect);
	}
	return MP_OK;
}

bool GetBaseAddresses( void )
{
	bool success = false;
	success = GetBaseAddress(SAMPLE_DLLFUNC, gameDllAddress);
	if(success == false) return false;

	success = GetBaseAddress(SAMPLE_ENGFUNC, gameEngAddress);
	if(success == false) return false;

	return true;
}

void ClientDisconnect(edict_t *pEntity)
{
	int index = ENTINDEX(pEntity);
	status[index][NAME_CHANGED] = false;
	RETURN_META(MRES_IGNORED);
}

void ClientUserInfoChanged(edict_t *pEntity, char *infobuffer)
{
	int index = ENTINDEX(pEntity);
	if (status[index][NAME_CHANGING])
	{
		status[index][NAME_CHANGING] = false;
		status[index][NAME_CHANGED] = true;
	}
	else
	{
		RETURN_META(MRES_IGNORED);
	}
	int i;
	int istrlen = (int)strlen(newname[index]);
	cell address = reinterpret_cast<cell>(strstr(infobuffer, "\\name\\"));
	cell realaddress = address + 6;
	for (i = 0; i<= istrlen; i++)
	{
		UTIL_PatchMemory_UnsignedByte(NULL, realaddress+i, newname[index][i], MEMTYPE_DATA);
	}
	RETURN_META(MRES_IGNORED);
}

static cell AMX_NATIVE_CALL ansi_to_utf8(AMX *amx, cell *params)
{
	int alen;
	int ulen;
	int unicodelen;
	char *ansi = MF_GetAmxString(amx, params[1], 0, &alen);
	if (!alen)
	{
		return 0;
	}
	WCHAR *unicodeBuffer;
	CHAR *utf8;
	//change ansi to unicode.
	unicodelen = MultiByteToWideChar(CP_ACP, NULL, ansi, alen, NULL, NULL);
	unicodeBuffer = new WCHAR[unicodelen];
	MultiByteToWideChar(CP_ACP, NULL, ansi, alen, unicodeBuffer, unicodelen);

	//change unicode to utf-8.
	ulen = WideCharToMultiByte(CP_UTF8, NULL, unicodeBuffer, unicodelen, NULL, NULL, NULL, NULL);
	utf8 = new CHAR[ulen];
	WideCharToMultiByte(CP_UTF8, NULL, unicodeBuffer, unicodelen, utf8, ulen, NULL, NULL);
	utf8[ulen] = '\0';
	MF_SetAmxString(amx, params[2], utf8, params[3]);
	delete [] unicodeBuffer;
	delete [] utf8;
	return 1;
}

static cell AMX_NATIVE_CALL utf8_to_ansi(AMX *amx, cell *params)
{
	int ulen;
	char *utf8 = MF_GetAmxString(amx, params[1], 0, &ulen);
	if (!ulen)
	{
		return 0;
	}
	int unicodelen;
	int alen;
	WCHAR *unicodeBuffer;
	char *ansi;

	unicodelen = MultiByteToWideChar(CP_UTF8, NULL, utf8, ulen, NULL, NULL);
	unicodeBuffer = new WCHAR[unicodelen];
	MultiByteToWideChar(CP_UTF8, NULL, utf8, ulen, unicodeBuffer, unicodelen);

	alen = WideCharToMultiByte(CP_ACP, NULL, unicodeBuffer, unicodelen, NULL, NULL, NULL, NULL);
	ansi = new char[alen];
	WideCharToMultiByte(CP_ACP, NULL, unicodeBuffer, unicodelen, ansi, alen, NULL, NULL);
	ansi[alen] = '\0';

	MF_SetAmxString(amx, params[2], ansi, params[3]);
	delete [] unicodeBuffer;
	delete [] ansi;
	return 1;
}
AMX_NATIVE_INFO my_natives[] = {
	{ "set_utf8_name",	set_utf8_name},
	{"ansi_to_utf8",	ansi_to_utf8},
	{"utf8_to_ansi",	utf8_to_ansi},
	{ NULL, NULL }
};

void OnAmxxAttach()
{
	if(GetBaseAddresses() == false) return MF_LogError(NULL,AMX_ERR_MEMACCESS,"Unable to get base address of game or mod .dll!");

	// Add natives if base is found.
	MF_AddNatives(my_natives);
	static char buf[1024];
	size_t len = 0;
	len = sprintf(buf, "==========Module Information==========\n\tName: %s\n\tVersion: %s\n", MODULE_NAME, MODULE_VERSION);
	len += sprintf(&buf[len], "\tCompiled: %s\n\tURL:%s\n=================End=================\n", __DATE__ ", " __TIME__, MODULE_URL);
	SERVER_PRINT(buf);
}